from copy import deepcopy
import numpy as np

class Collage():
    def __init__(self, key):
        self.key = key
        self.negative_space = None
        self.positive_space = None

    def __str__(self):
        return self.key

    def set_negative_space(self, custom_image):
        self.negative_space = custom_image

    def set_positive_space(self, custom_image):
        self.positive_space = custom_image

    def set_mask(self, mask_container):
        self.mask = mask_container

    def combine(self):
        #helpful post: https://stackoverflow.com/questions/46267443/merge-images-using-opencv-and-a-mask
        #making masks: https://codereview.stackexchange.com/questions/184044/processing-an-image-to-extract-green-screen-mask
        #creating alpha gradient image https://note.nkmk.me/en/python-numpy-generate-gradation-image/

        if len(self.mask.mask.shape) != 2:
            mask = self.mask.prepare_mask_for_collage_combine()#WHEN I COMMENT OUT 255, COLOR STAYS AT POSITIVE SPACE COLOR
            self.composite = (self.positive_space.image * mask + self.negative_space.image * (1 - self.mask.mask)).astype(np.uint8)
        #CREATES A composite image FROM boolean mask
        else:
            self.composite = copy(self.positive_space.image)
            self.composite[~self.mask.mask] = self.positive_space.image[~self.mask.mask]
            self.composite[self.mask.mask] = self.negative_space.image[self.mask.mask]#.astype(float))



from moonmask.collage import Collage
from moonmask.moon_image import MoonImage
from moonmask.custom_image import CustomImage
from moonmask.mask import Mask
from copy import deepcopy

class CollageGenerator():
    def __init__(self, size):
        self.collage_store = {}
        self.image_store = {} 
        self.mask_store = {} 
        self.selected_collage = None
        self.size = size

    def start_new_collage(self, keyword):
        self.collage_store[keyword] = Collage(keyword)
        self.selected_collage = self.collage_store[keyword]

    def add_new_image(self, key, image=None):
        self.image_store[key] = image 

    def add_new_mask(self, key, **kwargs):
        self.mask_store[key] = Mask(self.size, key, **kwargs) 

    def select_collage(self, key):
        self.selected_collage = self.collage_store[key]

    def load_new_image(self, key, **kwargs):
        self.image_store[key] = CustomImage(self.size, key, **kwargs)

    def duplicate_mask(self, source_key, destination_key):
        self.mask_store[destination_key] = deepcopy(self.mask_store[source_key])

    def duplicate_image(self, source_key, destination_key):
        self.image_store[destination_key] = deepcopy(self.image_store[source_key])

    def set_moon(self, key, **kwargs):
        """Sets the image that will be used as a mask on the image
        Keyword arguments:
        date -- the date in format YYYYMMDD
        relative_date -- defaults to "today". Other accepted options are "yesterday" and "tomorrow"
        """
        self.image_store[key] = MoonImage(self.size, key)
        self.image_store[key].set_moon_image(**kwargs)
        self.image_store[key]
        
        return

    def collage_to_image(self, collage_key, destination_image_key):
        self.load_new_image(destination_image_key, image_array=self.collage_store[collage_key].composite)

    def image_to_mask(self, source_image_key, destination_mask_key, **kwargs):
        self.add_new_mask(destination_mask_key, image=self.image_store[source_image_key].image, **kwargs)

 



import urllib.request
import io
import cv2
import numpy as np
from copy import deepcopy, copy

class CustomImage():
    def __init__(self, img_size, key, **kwargs):
        self.size = img_size
        self.key = key
        self.height = img_size[0]
        self.width = img_size[1]
        self.set_image(**kwargs)
        return

    def get_image(self):
        return self.image

    def set_image(self, image_array=None, url=None, instagram_url=None, filename=None, color=(0,0,0)):
        #if more than one of these parameters are provided, raise an error.

        if isinstance(image_array,np.ndarray): self.image = deepcopy(image_array)
        elif url: self.load_from_url(url)
        elif instagram_url: self.load_from_instagram(instagram_url)
        elif filename: self.load_from_filesystem(filename)
        elif color: self.make_custom_image(color)
        else: self.image = None
        return

    def make_custom_image(self, color):
        self.image = np.zeros((self.size[0],self.size[1],3), np.uint8)
        self.image[:,0:self.size[0]] = color
        return

    def add_vertical_stripe(self, beginning, end, color):
        beginning_index = int(beginning * self.width)
        end_index = int(end * self.width)

        #slicing numpy arrays
        self.image[:,beginning_index:end_index] = color
        return


    def add_horizontal_stripe(self, beginning, end, color):
        beginning_index = int(beginning * self.height)
        end_index = int(end * self.height)

        #slicing numpy arrays
        self.image[beginning_index:end_index,:] = color
        return

    def rot90(self, times=1):
        self.image = np.rot90(self.image, times)
        return

    def load_from_instagram(self, instagram_url):
        #TODO check if link has extra info at the end
        #https://nono.ma/says/get-an-instagram-image-url
        try:
            self.load_from_url(instagram_url + "media/?size=l")
        except:
            raise Exception("I can't seem to download that image from that link: " + instagram_url + ". Check that it's an instagram link in the form 'https://instagram.com/p/SOME_NUMBERS_AND_LETTERS/'.")
        return

    def load_from_filesystem(self, path):
        self.image = cv2.imread(path)
        self.resize_image()
        return

    def load_from_url(self, url):
        # METHOD #1: OpenCV, NumPy, and urllib
	    # download the image, convert it to a NumPy array, and then read
	    # it into OpenCV format
        # https://www.pyimagesearch.com/2015/03/02/convert-url-to-image-with-python-and-opencv/
        resp = urllib.request.urlopen(url)
        self.image = np.asarray(bytearray(resp.read()), dtype="uint8")
        self.image = cv2.imdecode(self.image, cv2.IMREAD_COLOR)
        print("resizing image")
        print(self.key)
        self.resize_image()
        return

    def resize_image(self):
        #https://medium.com/@manivannan_data/resize-image-using-opencv-python-d2cdbbc480f0
        #Preferable interpolation methods are cv.INTER_AREA for shrinking and cv.INTER_CUBIC(slow) & cv.INTER_LINEAR for zooming. By default, interpolation method used is cv.INTER_LINEAR for all resizing purposes.
        try:
            self.image = cv2.resize(self.image, self.size, interpolation=cv2.INTER_AREA)
        except Exception as e:
            print(str(e))
        self.image = cv2.resize(self.image, self.size, interpolation=cv2.INTER_CUBIC)
        print(self.image.shape)
        print(self.key)

    def pixelate_image(self, 
		       radius = 5, 
                       step = 14, 
                       jitter = 0, 
                       point_shape="circle",
                       radius_expansion_rate=1,
                       min_radius = 1,
                       max_radius = 100,
                       step_expansion_rate=1):
        height, width, channel = self.image.shape

        points = copy(self.image)

        for i in range(height):
            for j in range(width):
                points[i, j] = 255
        xrange = np.zeros(int(height/step))
        yrange = np.zeros(int(width/step))
        for xvalue in range(len(xrange)):
            xrange[xvalue] = xvalue
        for yvalue in range(len(yrange)):
            yrange[yvalue] = yvalue
        xrange = [value*step+step/2 for value in xrange]
        yrange= [value*step+step/2 for value in yrange]
        np.random.shuffle(xrange)
        for i in xrange:
            np.random.shuffle(yrange)
            for j in yrange:
                if radius * radius_expansion_rate >= min_radius and radius * radius_expansion_rate <= max_radius:
                    radius = int(radius * radius_expansion_rate)
                x = int(i)# + random.randint(1, 2*JITTER-JITTER))
                y = int(j)# + random.randint(1, 2*JITTER-JITTER))
                if(x >= height):
                        x = height-1
                if( y >= width):
                        y = width-1
                gray = self.image[x,y]
                gray = (int(gray[0]),int(gray[1]),int(gray[2]))
                if point_shape=="circle":
                    cv2.circle(points,
                            (y, x),
                            radius,
                            (gray),
                            -1,
                            cv2.LINE_AA)
                elif point_shape=="square":
                    cv2.rectangle(points,
                            (y, x),
                            (int(1 + y + radius/2), int(1 + x + radius/2)),
                            (gray),
                            -1,
                            cv2.LINE_AA)
        self.image = points
        return



import matplotlib
from moonmask.collage_generator import CollageGenerator
from moonmask.res.ascii_art import *
import numpy as np
import textwrap
import os
import cv2
import shutil
import math
import matplotlib.pyplot as plt

class JupyterUi(CollageGenerator):
    def __init__(self, size=(1000,1000)):
        super().__init__(size)
        self.gscale1 = "#########*oaZO0QLCJUYXzcvunxrjft/\|()1{}[]?-_+~<>i!lI;:,\"^`'.         "
        self.gscale2 = ' #@%#*+=-:. '

        self.terminal_width = shutil.get_terminal_size().columns
        self.text_width = int(self.terminal_width * 0.75)
        self.text_margin = int(self.terminal_width * 0.12)
        self.text_wrapper = textwrap.TextWrapper(width=self.text_width)
        self.margin_symbols = " "*(self.text_margin - 2)+"  "

        self.columns_per_image = 20 
        self.text_description_width = 20

    def preview(self):
        
        image_ascii_list = {} 
        for image_key in self.image_store:
            im_copy = self.image_store[image_key].image
            image_ascii_list[image_key] = self.convertImageToAscii(im_copy, self.columns_per_image, .5, self.gscale2)

        if len(image_ascii_list) > 0:
            self.print_ascii_art(IMAGES_TEXT)
            image_ascii_art_reordered = self.combine_ascii_images_on_line(image_ascii_list)
            self.print_ascii_art(image_ascii_art_reordered)

        collage_ascii_list = {}
        for image_key in self.collage_store:
            im_copy = self.collage_store[image_key].composite
            collage_ascii_list[image_key] = self.convertImageToAscii(im_copy, self.columns_per_image, .5, self.gscale2)
        
        if len(collage_ascii_list) > 0:
            self.print_ascii_art(COLLAGES_TEXT)
            collage_ascii_art_reordered = self.combine_ascii_images_on_line(collage_ascii_list)
            self.print_ascii_art(collage_ascii_art_reordered)

    def show_image(self, image_key):
        print("hi")
        plt.imshow(self.image_store[image_key].image)
        plt.show()

    def show_collage(self, collage_key):
        plt.imshow(self.collage_store[collage_key].composite)
        plt.show()

    def combine_ascii_images_on_line(self, image_ascii_list):
        image_ascii_formatted_list = []
        #self.text_width

        #we're going to make a list of lists
        #it will combine the 1st, 2nd, nth row of each image ascii art
        #each outlined by a border and with some textual description
        #on the right of the image

        #figure out how many images will fit on each line
        images_per_line = math.floor(self.text_width / (self.columns_per_image + self.text_description_width))

        #find out how many lines of images there will be
        lines_of_images = math.ceil(len(image_ascii_list) / images_per_line)
        
        #split dict of images into separate groups of # of images per line
        grouped_images_list = []
        image_keys = list(image_ascii_list.keys())
        num_rows=len(image_ascii_list[image_keys[0]])

        for x in range(lines_of_images):
            grouped_images_list.append([])
            for y in range(images_per_line):
                if len(image_keys) > 0:
                    grouped_images_list[x].append(image_keys.pop())
      
        grouped_images_dict = [] 
        for y in range(lines_of_images):
            grouped_images_dict.append({})
            for x in range(images_per_line):
                if x < len(grouped_images_list[y]):
                    grouped_images_dict[y][grouped_images_list[y][x]] = image_ascii_list[grouped_images_list[y][x]]
        image_names = []
        for group in range(len(grouped_images_dict)):
            image_names_line = []
            for image in grouped_images_dict[group]:
                image_names_line.append(image)
            image_names.append(image_names_line)
        
        ascii_image_lines = []
        for image_group_number in range(len(grouped_images_dict)):
            ascii_image_line = []
            for i in range(num_rows):
                terminal_line = [] 
                for image_key in grouped_images_dict[image_group_number]:
                    terminal_line.append(grouped_images_dict[image_group_number][image_key][i])
                ascii_image_line.append(terminal_line)
            ascii_image_lines.append(ascii_image_line)

        images_ascii_art = []
        for image_block_line in ascii_image_lines:
            for line in image_block_line:
                terminal_line = "                         ".join(line)
                images_ascii_art.append(terminal_line)
            images_ascii_art.append(" ")
        
        return images_ascii_art

        for x in image_names:
            x = (self.columns_per_image*" ").join(x)
            print(x) 
   
    def print_ascii_art(self, ascii_art_list):
        for element in ascii_art_list:
            self.print_text(element[:self.text_width])
        print("")

    def text_format(self,text):
        return textwrap.indent(text=self.text_wrapper.fill(text=text), prefix=self.margin_symbols)

    def print_list(self, list_to_print):
        for element in list_to_print:
            self.print_text((" * " + element))
            print("")

    def print_text(self, text):
        print(self.text_format(text))

    def print_multiline_text(self, text_list, substitute_text=""):
        for text in text_list:
            self.print_text(text.format(self.command_prompt + substitute_text))
            print("")

    def rgb_to_gray(self, img):
        grayImage = np.zeros(img.shape)
        R = np.array(img[:, :, 0])
        G = np.array(img[:, :, 1])
        B = np.array(img[:, :, 2])

        R = (R *.299)
        G = (G *.587)
        B = (B *.114)

        Avg = (R+G+B)
        grayImage = img

        for i in range(3):
           grayImage[:,:,i] = Avg

        return grayImage       

    ### PYTHON TO ASCII ART - https://github.com/electronut/pp/blob/master/ascii/ascii.py#L2 modified to just take numpy images

    def getAverageL(self, im):
        """
        Given PIL Image, return average value of grayscale value
        """
        # get shape
        w = im.shape[0]
        h = im.shape[1]
        # get average
        #cv2.imshow("test", im)
        #cv2.waitKey(0)
        #cv2.destroyAllWindows()
        return np.average(im)
    
    def convertImageToAscii(self, im, cols, scale, moreLevels):
        """
        Given Image and dims (rows, cols) returns an m*n list of Images 
        """
        # declare globals
        # open image and convert to grayscale
        im = self.rgb_to_gray(im)
        # store dimensions
        W, H = im.shape[0], im.shape[1]
        # compute width of tile
        w = W/cols
        # compute tile height based on aspect ratio and scale
        h = w/scale
        # compute number of rows
        rows = int(H/h)
        
        # check if image size is too small
        if cols > W or rows > H:
            print("Image too small for specified cols!")
            exit(0)
    
        # ascii image is a list of character strings
        aimg = []
        # generate list of dimensions
        for j in range(rows):
            y1 = int(j*h)
            y2 = int((j+1)*h)
            # correct last tile
            if j == rows-1:
                y2 = H
            # append an empty string
            aimg.append("")
            for i in range(cols):
                # crop image to tile
                x1 = int(i*w)
                x2 = int((i+1)*w)
                # correct last tile
                if i == cols-1:
                    x2 = W
                # crop image to extract tile
                #print(x1)
                #print(x2)
                #print(y1)
                #print(y2)
                img = im[x1:x2, y1:y2]
                #print(img)
                # get average luminance
                avg = int(self.getAverageL(img))
                # look up ascii char
                if moreLevels:
                    gsval = self.gscale1[int((avg*69)/255)]
                else:
                    gsval = self.gscale2[int((avg*9)/255)]
                # append ascii char to string
                aimg[j] += gsval
        
        # return txt image
        return aimg
   



import cv2
import numpy as np
from copy import copy
from bisect import bisect_left

class Mask():

    def __init__(self, size, key, image = [], alpha_values=0, mask_divisor=255,prep_mask_divisor=1):
        self.image = image
        self.key = key
        self.size = size
        self.mask_divisor = mask_divisor 
        self.prep_mask_divisor = prep_mask_divisor
        self.mask = self.set_mask(alpha_values) if len(image) > 0 else None

    def set_mask(self, alpha_values):
        """
        if alpha values is 0, all grey values in the image
        will be included in the mask
        """
        print("set mask method called")
        mask = self.rgb_to_gray(self.image)
        mask = mask / self.mask_divisor 
        if not alpha_values == 0:
            mask = self.normalize_mask_values(mask, alpha_values)
        return mask

    def prepare_mask_for_collage_combine(self):
        #WHEN PREP_MASK_DIVISOR IS 255, COLOR STAYS AT POSITIVE SPACE COLOR
        return self.mask / self.prep_mask_divisor

    def normalize_mask_values(self, mask, alpha_values):
        interval = 1.0 / (alpha_values - 1)
        values_list = []
        for x in range(alpha_values):
            values_list.append(x * interval) 
        for a in range(len(mask)):
            for b in range(len(mask[a])):
                for c in range(len(mask[a][b])):
                    mask[a][b][c] = self.take_closest(values_list, c)
        return mask


    def take_closest(self, sorted_list, number):
        """
        source: https://stackoverflow.com/a/12141511/5650506
        Assumes sorted_list is sorted. Returns closest value to number.
    
        If two numbers are equally close, return the smallest number.
        """
        pos = bisect_left(sorted_list, number)
        if pos == 0:
            return sorted_list[0]
        if pos == len(sorted_list):
            return sorted_list[-1]
        before = sorted_list[pos - 1]
        after = sorted_list[pos]
        if after - number < number - before:
           return after
        else:
           return before

    def rgb_to_gray(self, img):
        #https://stackoverflow.com/a/47380815/5650506
        gray_image = np.zeros(img.shape)
        R = np.array(img[:, :, 0])
        G = np.array(img[:, :, 1])
        B = np.array(img[:, :, 2])

        R = (R *.299)
        G = (G *.587)
        B = (B *.114)

        Avg = (R+G+B)
        grayImage = img

        for i in range(3):
           gray_image[:,:,i] = Avg

        return gray_image


    def pixelate_mask(self, radius = 15, step = 34, jitter = 0):
        #height, width, channel = self.image.shape

        #points = copy(self.image)

        #for i in range(height):
        #    for j in range(width):
        #        points[i, j] = 255
        #xrange = np.zeros(int(height/step))
        #yrange = np.zeros(int(width/step))
        #for xvalue in range(len(xrange)):
        #    xrange[xvalue] = xvalue
        #for yvalue in range(len(yrange)):
        #    yrange[yvalue] = yvalue
        #xrange = [value*step+step/2 for value in xrange]
        #yrange= [value*step+step/2 for value in yrange]
        #np.random.shuffle(xrange)
        #for i in xrange:
        #    np.random.shuffle(yrange)
        #    for j in yrange:
        #        x = int(i)# + random.randint(1, 2*JITTER-JITTER))
        #        y = int(j)# + random.randint(1, 2*JITTER-JITTER))
        #        if(x >= height):
        #                x = height-1
        #        if( y >= width):
        #                y = width-1
        #        gray = self.image[x,y]
        #        cv2.circle(points,
        #                (y, x),
        #                radius,
        #                int(gray),
        #                -1,
        #                cv2.LINE_AA)
        #self.pixelated_mask = points
        return

    def get_gradation_2d(self, start, stop, width, height, is_horizontal):
        #from https://note.nkmk.me/en/python-numpy-generate-gradation-image/
        if is_horizontal:
            return np.tile(np.linspace(start, stop, width), (height, 1))
        else:
            return np.tile(np.linspace(start, stop, height), (width, 1)).T

    def get_gradation_3d(self, width, height, start_list, stop_list, is_horizontal_list):
        #from https://note.nkmk.me/en/python-numpy-generate-gradation-image/
        result = np.zeros((height, width, len(start_list)), dtype=np.float)

        for i, (start, stop, is_horizontal) in enumerate(zip(start_list, stop_list, is_horizontal_list)):
            result[:, :, i] = self.get_gradation_2d(start, stop, width, height, is_horizontal)

        return result





from distutils.core import setup
from setuptools import setup, find_packages

setup(
    name = 'moonmask',
    packages = find_packages(),
    version = '2.2',
    license='MIT',
    description = 'Gets moon visualizations courtesy of Ernie Wright and uses them for artistic collages',
    author = 'Sadie Parker',
    author_email = 'sadiemparker@gmail.com',
    url = 'https://github.com/spacerest/moonmask',
    download_url = 'https://github.com/spacerest/moonmask/archive/v_22.tar.gz',
    keywords = ['MOON', 'ART'],
    install_requires=[
        'atomicwrites==1.3.0',
        'attrs==19.1.0',
        'importlib-metadata==0.18',
        'more-itertools==7.2.0',
        'numpy==1.16.4',
        'opencv-python==4.1.0.25',
        'packaging==19.0',
        'pluggy==0.12.0',
        'py==1.8.0',
        'pyparsing==2.4.2',
        'pytest==5.0.1',
        'six==1.12.0',
        'wcwidth==0.1.7',
        'zipp==0.5.2',
    ],
    classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Other Audience',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
